#ifndef __SYNC_CRITICAL_RESOURCE_HXX_INCLUDED__
#define __SYNC_CRITICAL_RESOURCE_HXX_INCLUDED__

#include "SyncCriticalResource.h"

namespace sync
{

// class TResourceLocker

template<typename TResourceType>
TCriticalResource<TResourceType>::TResourceLocker::TResourceLocker(TResourceType& resource,
                                                                 sync::CCriticalSection& access
                                                                )
: _resource(&resource)
, _access(&access)
{
   _access->enter();
}

template<typename TResourceType>
TCriticalResource<TResourceType>::TResourceLocker::TResourceLocker
   (const TResourceLocker& resource)
: _resource(resource._resource)
, _access(resource._access)
{
   _access->enter();
}

template<typename TResourceType>
typename TCriticalResource<TResourceType>::TResourceLocker&
   TCriticalResource<TResourceType>::TResourceLocker::operator=(const TResourceType& resource)
{
   (*_resource) = resource;
   return (*this);
}

template<typename TResourceType>
TCriticalResource<TResourceType>::TResourceLocker::~TResourceLocker() throw()
{
   _access->leave();
}

template<typename TResourceType>
TCriticalResource<TResourceType>::TResourceLocker::operator TResourceType&() const
{
   return (*_resource);
}

template<typename TResourceType>
TResourceType& TCriticalResource<TResourceType>::TResourceLocker::get() const
{
   return (*_resource);
}

template<typename TResourceType>
typename TCriticalResource<TResourceType>::TResourceLocker*
TCriticalResource<TResourceType>::TResourceLocker::operator->()
{
   return this;
}

template<typename TResourceType>
typename TCriticalResource<TResourceType>::TResourceLocker*
TCriticalResource<TResourceType>::TResourceLocker::operator*()
{
   return this;
}

// class TConstResourceLocker

template<typename TResourceType>
TCriticalResource<TResourceType>::TConstResourceLocker::TConstResourceLocker(const TResourceType& resource,
                                                                           sync::CCriticalSection& access
                                                                          )
: _resource(&resource)
, _access(&access)
{
   _access->enter();
}

template<typename TResourceType>
TCriticalResource<TResourceType>::TConstResourceLocker::TConstResourceLocker
   (const TConstResourceLocker& resource)
: _resource(resource._resource)
, _access(resource._access)
{
   _access->enter();
}

template<typename TResourceType>
TCriticalResource<TResourceType>::TConstResourceLocker::~TConstResourceLocker() throw()
{
   _access->leave();
}

template<typename TResourceType>
TCriticalResource<TResourceType>::TConstResourceLocker::operator const TResourceType&() const
{
   return (*_resource);
}

template<typename TResourceType>
const TResourceType& TCriticalResource<TResourceType>::TConstResourceLocker::get() const
{
   return (*_resource);
}

template<typename TResourceType>
typename TCriticalResource<TResourceType>::TConstResourceLocker*
TCriticalResource<TResourceType>::TConstResourceLocker::operator->()
{
   return this;
}

template<typename TResourceType>
typename TCriticalResource<TResourceType>::TConstResourceLocker*
TCriticalResource<TResourceType>::TConstResourceLocker::operator*()
{
   return this;
}

// class TCriticalResource

template<typename TResourceType>
TCriticalResource<TResourceType>::TCriticalResource()
: _resource()
, _access()
{
}

template<typename TResourceType>
TCriticalResource<TResourceType>::TCriticalResource(const TResourceType& resource)
: _resource(resource)
, _access()
{
}

template<typename TResourceType>
TCriticalResource<TResourceType>::TCriticalResource(const TCriticalResource<TResourceType>& resource)
: _resource(resource._resource)
, _access()
{
}

template<typename TResourceType>
TCriticalResource<TResourceType>& TCriticalResource<TResourceType>::operator=
   (const TCriticalResource<TResourceType>& resource)
{  // Resource copy user in operator= so do not enter in more that one critical section at once.
   if (this != &resource)
   {
      TResourceType resourceCopy = resource.copy();
      (*this)->get() = resourceCopy;
   }
   return (*this);
}

template<typename TResourceType>
typename TCriticalResource<TResourceType>::TResourceLocker
TCriticalResource<TResourceType>::operator->()
{
   return TCriticalResource<TResourceType>::TResourceLocker(_resource, _access);
}

template<typename TResourceType>
typename TCriticalResource<TResourceType>::TConstResourceLocker
TCriticalResource<TResourceType>::operator->() const
{
   return TCriticalResource<TResourceType>::TConstResourceLocker(_resource, _access);
}

template<typename TResourceType>
typename TCriticalResource<TResourceType>::TResourceLocker
TCriticalResource<TResourceType>::operator*()
{
   return TCriticalResource<TResourceType>::TResourceLocker(_resource, _access);
}

template<typename TResourceType>
typename TCriticalResource<TResourceType>::TConstResourceLocker
TCriticalResource<TResourceType>::operator*() const
{
   return TCriticalResource<TResourceType>::TConstResourceLocker(_resource, _access);
}

template<typename TResourceType>
TResourceType TCriticalResource<TResourceType>::copy() const
{
   return (*this)->get();
}

template<typename TResourceType>
sync::CCriticalSection& TCriticalResource<TResourceType>::section()
{
   return _access;
}

//
// References specialization
//

// class TCriticalResource<TResourceType&>::TResourceLocker

template<typename TResourceType>
TCriticalResource<TResourceType&>::TResourceLocker::TResourceLocker(TResourceType& resource,
                                                                  sync::CCriticalSection& access
                                                                 )
: _resource(&resource)
, _access(&access)
{
   _access->enter();
}

template<typename TResourceType>
TCriticalResource<TResourceType&>::TResourceLocker::TResourceLocker
   (const TResourceLocker& resource)
: _resource(resource._resource)
, _access(resource._access)
{
   _access->enter();
}

template<typename TResourceType>
TCriticalResource<TResourceType&>::TConstResourceLocker::~TConstResourceLocker() throw()
{
   _access->leave();
}

template<typename TResourceType>
typename TCriticalResource<TResourceType&>::TResourceLocker&
   TCriticalResource<TResourceType&>::TResourceLocker::operator=(const TResourceType& resource)
{
   (*_resource) = resource;
   return (*this);
}

template<typename TResourceType>
TCriticalResource<TResourceType&>::TResourceLocker::~TResourceLocker() throw()
{
   _access->leave();
}

template<typename TResourceType>
TCriticalResource<TResourceType&>::TResourceLocker::operator TResourceType&() const
{
   return (*_resource);
}

template<typename TResourceType>
TResourceType& TCriticalResource<TResourceType&>::TResourceLocker::get() const
{
   return (*_resource);
}

template<typename TResourceType>
typename TCriticalResource<TResourceType&>::TResourceLocker*
TCriticalResource<TResourceType&>::TResourceLocker::operator->()
{
   return this;
}

template<typename TResourceType>
typename TCriticalResource<TResourceType&>::TResourceLocker*
TCriticalResource<TResourceType&>::TResourceLocker::operator*()
{
   return this;
}

// class TCriticalResource<TResourceType&>::TConstResourceLocker

template<typename TResourceType>
TCriticalResource<TResourceType&>::TConstResourceLocker::TConstResourceLocker(const TResourceType& resource,
                                                                            sync::CCriticalSection& access
                                                                           )
: _resource(&resource)
, _access(&access)
{
   _access->enter();
}

template<typename TResourceType>
TCriticalResource<TResourceType&>::TConstResourceLocker::TConstResourceLocker
   (const TConstResourceLocker& resource)
: _resource(resource._resource)
, _access(resource._access)
{
   _access->enter();
}

template<typename TResourceType>
TCriticalResource<TResourceType&>::TConstResourceLocker::operator const TResourceType&() const
{
   return (*_resource);
}

template<typename TResourceType>
const TResourceType& TCriticalResource<TResourceType&>::TConstResourceLocker::get() const
{
   return (*_resource);
}

template<typename TResourceType>
typename TCriticalResource<TResourceType&>::TConstResourceLocker*
TCriticalResource<TResourceType&>::TConstResourceLocker::operator->()
{
   return this;
}

template<typename TResourceType>
typename TCriticalResource<TResourceType&>::TConstResourceLocker*
TCriticalResource<TResourceType&>::TConstResourceLocker::operator*()
{
   return this;
}

// class TCriticalResource<TResourceType&>

template<typename TResourceType>
TCriticalResource<TResourceType&>::TCriticalResource(TResourceType& resource)
: _resource(&resource)
, _access()
{
}

template<typename TResourceType>
typename TCriticalResource<TResourceType&>::TResourceLocker
TCriticalResource<TResourceType&>::operator->()
{
   return TCriticalResource<TResourceType&>::TResourceLocker(*_resource, _access);
}

template<typename TResourceType>
typename TCriticalResource<TResourceType&>::TConstResourceLocker
TCriticalResource<TResourceType&>::operator->() const
{
   return TCriticalResource<TResourceType&>::TConstResourceLocker(*_resource, _access);
}

template<typename TResourceType>
typename TCriticalResource<TResourceType&>::TResourceLocker
TCriticalResource<TResourceType&>::operator*()
{
   return TCriticalResource<TResourceType&>::TResourceLocker(*_resource, _access);
}

template<typename TResourceType>
typename TCriticalResource<TResourceType&>::TConstResourceLocker
TCriticalResource<TResourceType&>::operator*() const
{
   return TCriticalResource<TResourceType&>::TConstResourceLocker(*_resource, _access);
}

template<typename TResourceType>
TResourceType TCriticalResource<TResourceType&>::copy() const
{
   return (*this)->get();
}

template<typename TResourceType>
sync::CCriticalSection& TCriticalResource<TResourceType&>::section()
{
   return _access;
}

} // namespace sync

#endif // __SYNC_CRITICAL_RESOURCE_HXX_INCLUDED__